@@ -1,7 +1,7 @@
 /* -*- c++ -*- (enables emacs c++ mode) */
 /*===========================================================================
 
- Copyright (C) 1999-2016 Yves Renard
+ Copyright (C) 1999-2017 Yves Renard
 
  This file is a part of GetFEM++
 
@@ -88,23 +88,29 @@
    - "FEM_REDUCED_QUADC1_COMPOSITE" : quadrilateral element, composite
       P3 element and C^1 (12 dof).
 
-   - "FEM_PK_HIERARCHICAL(N,K)" : PK element with a hierarchical basis
+   - "FEM_PK_HIERARCHICAL(N,K)" : PK element with a hierarchical basis.
 
-   - "FEM_QK_HIERARCHICAL(N,K)" : QK element with a hierarchical basis
+   - "FEM_QK_HIERARCHICAL(N,K)" : QK element with a hierarchical basis.
 
    - "FEM_PK_PRISM_HIERARCHICAL(N,K)" : PK element on a prism with a
-   hierarchical basis
+   hierarchical basis.
 
    - "FEM_STRUCTURED_COMPOSITE(FEM, K)" : Composite fem on a grid with
-   K divisions
+   K divisions.
 
    - "FEM_PK_HIERARCHICAL_COMPOSITE(N,K,S)" : PK composite element on
-   a grid with S subdivisions and with a hierarchical basis
+   a grid with S subdivisions and with a hierarchical basis.
 
    - "FEM_PK_FULL_HIERARCHICAL_COMPOSITE(N,K,S)" : PK composite
    element with S subdivisions and a hierarchical basis on both degree
-   and subdivision
+   and subdivision.
 
+   - "FEM_PYRAMID_LAGRANGE(K)" : Lagrange element on a 3D pyramid of degree
+   K=0, 1 or 2. Can be connected to a standard P1/P2 lagrange on its
+   triangular faces and standard Q1/Q2 Lagrange on its quadrangular face.
+
+   - "FEM_PYRAMID_DISCONTINUOUS_LAGRANGE(K)" : Discontinuous Lagrange element
+   on a 3D pyramid of degree K = 0, 1 or 2.
 */
 
 #ifndef GETFEM_FEM_H__
@@ -253,6 +259,7 @@ namespace getfem {
     bool is_pol;            // true if the FEM is polynomial
     bool is_polycomp;       // true if the FEM is polynomial composite
     bool real_element_defined;
+    bool is_standard_fem;   // is_equiv && !real_element_defined && scalar
     short_type es_degree;   // estimated polynomial degree of the FEM
     short_type hier_raff;   // hierarchical refinement of the FEM
     vec_type vtype; // for vectorial elements, type of transformation
@@ -328,10 +335,12 @@ namespace getfem {
     /// true if the base functions are polynomials
     bool is_polynomial() const { return is_pol; }
     bool is_polynomialcomp() const { return is_polycomp; }
+    bool is_standard() const { return is_standard_fem; }
     bool &is_polynomialcomp() { return is_polycomp; }
     bool &is_equivalent() { return is_equiv; }
     bool &is_lagrange() { return is_lag; }
     bool &is_polynomial() { return is_pol; }
+    bool &is_standard() { return is_standard_fem; }
     short_type estimated_degree() const { return es_degree; }
     short_type &estimated_degree() { return es_degree; }
 
@@ -453,7 +462,7 @@ namespace getfem {
     virtual_fem() {
       DAL_STORED_OBJECT_DEBUG_CREATED(this, "Fem");
       ntarget_dim = 1; dim_ = 1;
-      is_equiv = is_pol = is_polycomp = is_lag = false;
+      is_equiv = is_pol = is_polycomp = is_lag = is_standard_fem = false;
       pspt_valid = false; hier_raff = 0; real_element_defined = false;
       es_degree = 5;
       vtype = VECTORIAL_NOTRANSFORM_TYPE;
@@ -476,7 +485,35 @@ namespace getfem {
   template <class FUNC> class fem : public virtual_fem {
   protected :
     std::vector<FUNC> base_;
+    mutable std::vector<std::vector<FUNC>> grad_, hess_;
 
+    void compute_grad_() const {
+      size_type R = nb_base_components(0);
+      dim_type n = dim();
+      grad_.resize(R);
+      for (size_type i = 0; i < R; ++i) {
+	grad_[i].resize(n);
+	for (dim_type j = 0; j < n; ++j) {
+	  grad_[i][j] = base_[i]; grad_[i][j].derivative(j);
+	}
+      }
+    }
+
+    void compute_hess_() const {
+      size_type R = nb_base_components(0);
+      dim_type n = dim();
+      hess_.resize(R);
+      for (size_type i = 0; i < R; ++i) {
+	hess_[i].resize(n*n);
+	for (dim_type j = 0; j < n; ++j) {
+	  for (dim_type k = 0; k < n; ++k) {
+	    hess_[i][j+k*n] = base_[i];
+	    hess_[i][j+k*n].derivative(j); hess_[i][j+k*n].derivative(k);
+	  }
+	}
+      }
+    }
+    
   public :
 
     /// Gives the array of basic functions (components).
@@ -497,6 +534,7 @@ namespace getfem {
         reference element directions 0,..,dim-1 and returns the result in
         t(nb_base,target_dim,dim) */
     void grad_base_value(const base_node &x, base_tensor &t) const {
+      if (!(grad_.size())) compute_grad_();
       bgeot::multi_index mi(3);
       dim_type n = dim();
       mi[2] = n; mi[1] = target_dim(); mi[0] = short_type(nb_base(0));
@@ -505,12 +543,13 @@ namespace getfem {
       base_tensor::iterator it = t.begin();
       for (dim_type j = 0; j < n; ++j)
         for (size_type i = 0; i < R; ++i, ++it)
-          { FUNC f = base_[i]; f.derivative(j); *it = bgeot::to_scalar(f.eval(x.begin())); }
+	  *it = bgeot::to_scalar(grad_[i][j].eval(x.begin()));
     }
     /** Evaluates at point x, the hessian of all base functions w.r.t. the
         reference element directions 0,..,dim-1 and returns the result in
         t(nb_base,target_dim,dim,dim) */
     void hess_base_value(const base_node &x, base_tensor &t) const {
+      if (!(hess_.size())) compute_hess_();
       bgeot::multi_index mi(4);
       dim_type n = dim();
       mi[3] = n; mi[2] = n; mi[1] = target_dim();
@@ -520,19 +559,18 @@ namespace getfem {
       base_tensor::iterator it = t.begin();
       for (dim_type k = 0; k < n; ++k)
         for (dim_type j = 0; j < n; ++j)
-          for (size_type i = 0; i < R; ++i, ++it) {
-            FUNC f = base_[i];
-            f.derivative(j); f.derivative(k);
-            *it = bgeot::to_scalar(f.eval(x.begin()));
-          }
+          for (size_type i = 0; i < R; ++i, ++it)
+	    *it = bgeot::to_scalar(hess_[i][j+k*n].eval(x.begin()));
     }
 
   };
 
   /** Classical polynomial FEM. */
-  typedef const fem<base_poly> * ppolyfem;
+  typedef const fem<bgeot::base_poly> * ppolyfem;
   /** Polynomial composite FEM */
   typedef const fem<bgeot::polynomial_composite> * ppolycompfem;
+  /** Rational fration FEM */
+  typedef const fem<bgeot::base_rational_fraction> * prationalfracfem;
 
   /** Give a pointer on the structures describing the classical
       polynomial fem of degree k on a given convex type.
@@ -686,10 +724,14 @@ namespace getfem {
         at point @c this->xref())
     */
     void base_value(base_tensor& t, bool withM = true) const;
+    // Optimized function for high level generic assembly
+    void pfp_base_value(base_tensor& t, const pfem_precomp &pfp__);
     /** fill the tensor with the gradient of the base functions (taken
         at point @c this->xref())
     */
     void grad_base_value(base_tensor& t, bool withM = true) const;
+    // Optimized function for high level generic assembly
+    void pfp_grad_base_value(base_tensor& t, const pfem_precomp &pfp__);
     /** fill the tensor with the hessian of the base functions (taken
         at point @c this->xref())
     */
@@ -699,6 +741,7 @@ namespace getfem {
     /** get the current convex number */
     size_type convex_num() const;
     bool is_convex_num_valid() const;
+    void invalid_convex_num() { convex_num_ = size_type(-1); }
     /** set the current face number */
     void set_face_num(short_type f) { face_num_ = f; }
     /** get the current face number */
@@ -711,23 +754,59 @@ namespace getfem {
     void set_pf(pfem newpf);
     int xfem_side() const { return xfem_side_; }
     void set_xfem_side(int side) { xfem_side_ = side; }
-    fem_interpolation_context();
+    void change(bgeot::pgeotrans_precomp pgp__,
+		pfem_precomp pfp__, size_type ii__,
+		const base_matrix& G__, size_type convex_num__,
+		short_type face_num__ = short_type(-1)) {
+      bgeot::geotrans_interpolation_context::change(pgp__,ii__,G__);
+      convex_num_ = convex_num__; face_num_ = face_num__;  xfem_side_ = 0;
+      set_pfp(pfp__);
+    }
+    void change(bgeot::pgeometric_trans pgt__,
+		pfem_precomp pfp__, size_type ii__,
+		const base_matrix& G__, size_type convex_num__,
+		short_type face_num__ = short_type(-1)) {
+      bgeot::geotrans_interpolation_context::change
+	(pgt__, pfp__->get_ppoint_tab(), ii__, G__);
+      convex_num_ = convex_num__; face_num_ = face_num__; xfem_side_ = 0;
+      set_pfp(pfp__);
+    }
+    void change(bgeot::pgeometric_trans pgt__,
+		pfem pf__, const base_node& xref__, const base_matrix& G__,
+		size_type convex_num__,	short_type face_num__=short_type(-1)) {
+      bgeot::geotrans_interpolation_context::change(pgt__,xref__,G__);
+      pf_ = pf__; pfp_ = 0; convex_num_ = convex_num__; face_num_ = face_num__;
+      xfem_side_ = 0;
+    }
+    fem_interpolation_context()
+      : bgeot::geotrans_interpolation_context(),
+	convex_num_(size_type(-1)), face_num_(short_type(-1)), xfem_side_(0) {}
     fem_interpolation_context(bgeot::pgeotrans_precomp pgp__,
                               pfem_precomp pfp__, size_type ii__,
                               const base_matrix& G__,
                               size_type convex_num__,
-                              short_type face_num__ = short_type(-1));
+                              short_type face_num__ = short_type(-1))
+      : bgeot::geotrans_interpolation_context(pgp__,ii__,G__),
+	convex_num_(convex_num__), face_num_(face_num__), xfem_side_(0)
+    { set_pfp(pfp__); }
     fem_interpolation_context(bgeot::pgeometric_trans pgt__,
                               pfem_precomp pfp__, size_type ii__,
                               const base_matrix& G__,
                               size_type convex_num__,
-                              short_type face_num__ = short_type(-1));
+                              short_type face_num__ = short_type(-1))
+      : bgeot::geotrans_interpolation_context(pgt__,pfp__->get_ppoint_tab(),
+					      ii__, G__),
+	convex_num_(convex_num__), face_num_(face_num__), xfem_side_(0)
+    { set_pfp(pfp__); }
     fem_interpolation_context(bgeot::pgeometric_trans pgt__,
                               pfem pf__,
                               const base_node& xref__,
                               const base_matrix& G__,
                               size_type convex_num__,
-                              short_type face_num__ = short_type(-1));
+                              short_type face_num__ = short_type(-1))
+      : bgeot::geotrans_interpolation_context(pgt__,xref__,G__),
+	pf_(pf__), pfp_(0), convex_num_(convex_num__), face_num_(face_num__),
+	xfem_side_(0) {}
   };
 
   // IN : coeff(Qmult,nb_dof)
